@@ -6,6 +6,7 @@ import hmac
 import logging
 import math
 import secrets
+import string
 import time
 
 import aiohttp
@@ -90,39 +91,6 @@ class RoborockApiClient:
         md5.update(self._device_identifier.encode())
         return base64.b64encode(md5.digest()).decode()
 
-    def _process_extra_hawk_values(self, values: dict | None) -> str:
-        if values is None:
-            return ""
-        else:
-            sorted_keys = sorted(values.keys())
-            result = []
-            for key in sorted_keys:
-                value = values.get(key)
-                result.append(f"{key}={value}")
-            return hashlib.md5("&".join(result).encode()).hexdigest()
-
-    def _get_hawk_authentication(
-        self, rriot: RRiot, url: str, formdata: dict | None = None, params: dict | None = None
-    ) -> str:
-        timestamp = math.floor(time.time())
-        nonce = secrets.token_urlsafe(6)
-        formdata_str = self._process_extra_hawk_values(formdata)
-        params_str = self._process_extra_hawk_values(params)
-
-        prestr = ":".join(
-            [
-                rriot.u,
-                rriot.s,
-                nonce,
-                str(timestamp),
-                hashlib.md5(url.encode()).hexdigest(),
-                params_str,
-                formdata_str,
-            ]
-        )
-        mac = base64.b64encode(hmac.new(rriot.h.encode(), prestr.encode(), hashlib.sha256).digest()).decode()
-        return f'Hawk id="{rriot.u}",s="{rriot.s}",ts="{timestamp}",nonce="{nonce}",mac="{mac}"'
-
     async def nc_prepare(self, user_data: UserData, timezone: str) -> dict:
         """This gets a few critical parameters for adding a device to your account."""
         if (
@@ -144,7 +112,7 @@ class RoborockApiClient:
             "post",
             "/nc/prepare",
             headers={
-                "Authorization": self._get_hawk_authentication(
+                "Authorization": _get_hawk_authentication(
                     user_data.rriot, "/nc/prepare", {"hid": hid, "tzid": timezone}
                 ),
             },
@@ -177,7 +145,7 @@ class RoborockApiClient:
             "GET",
             "/user/devices/newadd",
             headers={
-                "Authorization": self._get_hawk_authentication(
+                "Authorization": _get_hawk_authentication(
                     user_data.rriot, "/user/devices/newadd", params={"s": s, "t": t}
                 ),
             },
@@ -223,6 +191,113 @@ class RoborockApiClient:
             else:
                 raise RoborockException(f"{code_response.get('msg')} - response code: {code_response.get('code')}")
 
+    async def request_code_v4(self) -> None:
+        """Request a code using the v4 endpoint."""
+        try:
+            self._login_limiter.try_acquire("login")
+        except BucketFullException as ex:
+            _LOGGER.info(ex.meta_info)
+            raise RoborockRateLimit("Reached maximum requests for login. Please try again later.") from ex
+        base_url = await self._get_base_url()
+        header_clientid = self._get_header_client_id()
+        code_request = PreparedRequest(
+            base_url,
+            self.session,
+            {
+                "header_clientid": header_clientid,
+                "Content-Type": "application/x-www-form-urlencoded",
+                "header_clientlang": "en",
+            },
+        )
+
+        code_response = await code_request.request(
+            "post",
+            "/api/v4/email/code/send",
+            params={"email": self._username, "type": "login", "platform": ""},
+        )
+        if code_response is None:
+            raise RoborockException("Failed to get a response from send email code")
+        response_code = code_response.get("code")
+        if response_code != 200:
+            _LOGGER.info("Request code failed for %s with the following context: %s", self._username, code_response)
+            if response_code == 2008:
+                raise RoborockAccountDoesNotExist("Account does not exist - check your login and try again.")
+            elif response_code == 9002:
+                raise RoborockTooFrequentCodeRequests("You have attempted to request too many codes. Try again later")
+            else:
+                raise RoborockException(f"{code_response.get('msg')} - response code: {code_response.get('code')}")
+
+    async def sign_key_v3(self, s: str) -> str:
+        """Sign a randomly generated string."""
+        base_url = await self._get_base_url()
+        header_clientid = self._get_header_client_id()
+        code_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
+
+        code_response = await code_request.request(
+            "post",
+            "/api/v3/key/sign",
+            params={"s": s},
+        )
+
+        if not code_response or "data" not in code_response or "k" not in code_response["data"]:
+            raise RoborockException("Failed to get a response from sign key")
+        response_code = code_response.get("code")
+
+        if response_code != 200:
+            _LOGGER.info("Request code failed for %s with the following context: %s", self._username, code_response)
+            raise RoborockException(f"{code_response.get('msg')} - response code: {code_response.get('code')}")
+
+        return code_response["data"]["k"]
+
+    async def code_login_v4(self, code: int | str, country: str, country_code: int) -> UserData:
+        """
+        Login via code authentication.
+        :param code: The code from the email.
+        :param country: The two-character representation of the country, i.e. "US"
+        :param country_code: the country phone number code i.e. 1 for US.
+        """
+        base_url = await self._get_base_url()
+        header_clientid = self._get_header_client_id()
+        x_mercy_ks = "".join(secrets.choice(string.ascii_letters + string.digits) for _ in range(16))
+        x_mercy_k = await self.sign_key_v3(x_mercy_ks)
+        login_request = PreparedRequest(
+            base_url,
+            self.session,
+            {"header_clientid": header_clientid, "x-mercy-ks": x_mercy_ks, "x-mercy-k": x_mercy_k},
+        )
+        login_response = await login_request.request(
+            "post",
+            "/api/v4/auth/email/login/code",
+            params={
+                "country": country,
+                "countryCode": country_code,
+                "email": self._username,
+                "code": code,
+                # Major and minor version are the user agreement version, we will need to see if this needs to be
+                # dynamic https://usiot.roborock.com/api/v3/app/agreement/latest?country=US
+                "majorVersion": 14,
+                "minorVersion": 0,
+            },
+        )
+        if login_response is None:
+            raise RoborockException("Login request response is None")
+        response_code = login_response.get("code")
+        if response_code != 200:
+            _LOGGER.info("Login failed for %s with the following context: %s", self._username, login_response)
+            if response_code == 2018:
+                raise RoborockInvalidCode("Invalid code - check your code and try again.")
+            if response_code == 3009:
+                raise RoborockNoUserAgreement("You must accept the user agreement in the Roborock app to continue.")
+            if response_code == 3006:
+                raise RoborockInvalidUserAgreement(
+                    "User agreement must be accepted again - or you are attempting to use the Mi Home app account."
+                )
+            raise RoborockException(f"{login_response.get('msg')} - response code: {response_code}")
+        user_data = login_response.get("data")
+        if not isinstance(user_data, dict):
+            raise RoborockException("Got unexpected data type for user_data")
+        return UserData.from_dict(user_data)
+
     async def pass_login(self, password: str) -> UserData:
         try:
             self._login_limiter.try_acquire("login")
@@ -337,7 +412,7 @@ class RoborockApiClient:
             rriot.r.a,
             self.session,
             {
-                "Authorization": self._get_hawk_authentication(rriot, f"/user/homes/{str(home_id)}"),
+                "Authorization": _get_hawk_authentication(rriot, f"/user/homes/{str(home_id)}"),
             },
         )
         home_response = await home_request.request("get", "/user/homes/" + str(home_id))
@@ -366,7 +441,7 @@ class RoborockApiClient:
             rriot.r.a,
             self.session,
             {
-                "Authorization": self._get_hawk_authentication(rriot, "/v2/user/homes/" + str(home_id)),
+                "Authorization": _get_hawk_authentication(rriot, "/v2/user/homes/" + str(home_id)),
             },
         )
         home_response = await home_request.request("get", "/v2/user/homes/" + str(home_id))
@@ -393,7 +468,7 @@ class RoborockApiClient:
             rriot.r.a,
             self.session,
             {
-                "Authorization": self._get_hawk_authentication(rriot, "/v3/user/homes/" + str(home_id)),
+                "Authorization": _get_hawk_authentication(rriot, "/v3/user/homes/" + str(home_id)),
             },
         )
         home_response = await home_request.request("get", "/v3/user/homes/" + str(home_id))
@@ -416,7 +491,7 @@ class RoborockApiClient:
             rriot.r.a,
             self.session,
             {
-                "Authorization": self._get_hawk_authentication(rriot, "/v2/user/homes/" + str(home_id)),
+                "Authorization": _get_hawk_authentication(rriot, "/v2/user/homes/" + str(home_id)),
             },
         )
         room_response = await room_request.request("get", f"/user/homes/{str(home_id)}/rooms" + str(home_id))
@@ -441,7 +516,7 @@ class RoborockApiClient:
             rriot.r.a,
             self.session,
             {
-                "Authorization": self._get_hawk_authentication(rriot, f"/user/scene/device/{str(device_id)}"),
+                "Authorization": _get_hawk_authentication(rriot, f"/user/scene/device/{str(device_id)}"),
             },
         )
         scenes_response = await scenes_request.request("get", f"/user/scene/device/{str(device_id)}")
@@ -463,7 +538,7 @@ class RoborockApiClient:
             rriot.r.a,
             self.session,
             {
-                "Authorization": self._get_hawk_authentication(rriot, f"/user/scene/{str(scene_id)}/execute"),
+                "Authorization": _get_hawk_authentication(rriot, f"/user/scene/{str(scene_id)}/execute"),
             },
         )
         execute_scene_response = await execute_scene_request.request("POST", f"/user/scene/{str(scene_id)}/execute")
@@ -546,3 +621,36 @@ class PreparedRequest:
         finally:
             if close_session:
                 await session.close()
+
+
+def _process_extra_hawk_values(values: dict | None) -> str:
+    if values is None:
+        return ""
+    else:
+        sorted_keys = sorted(values.keys())
+        result = []
+        for key in sorted_keys:
+            value = values.get(key)
+            result.append(f"{key}={value}")
+        return hashlib.md5("&".join(result).encode()).hexdigest()
+
+
+def _get_hawk_authentication(rriot: RRiot, url: str, formdata: dict | None = None, params: dict | None = None) -> str:
+    timestamp = math.floor(time.time())
+    nonce = secrets.token_urlsafe(6)
+    formdata_str = _process_extra_hawk_values(formdata)
+    params_str = _process_extra_hawk_values(params)
+
+    prestr = ":".join(
+        [
+            rriot.u,
+            rriot.s,
+            nonce,
+            str(timestamp),
+            hashlib.md5(url.encode()).hexdigest(),
+            params_str,
+            formdata_str,
+        ]
+    )
+    mac = base64.b64encode(hmac.new(rriot.h.encode(), prestr.encode(), hashlib.sha256).digest()).decode()
+    return f'Hawk id="{rriot.u}",s="{rriot.s}",ts="{timestamp}",nonce="{nonce}",mac="{mac}"'
